@@ -7,50 +7,51 @@ linking and assembling.
 import io
 import logging
 import os
+import sys
 import platform
 import stat
 import xml
 from .arch.arch import Architecture
-from .lang.c import CBuilder, CPreProcessor, COptions
-from .lang.c.preprocessor import CTokenPrinter
+from .lang.c import preprocess, c_to_ir, COptions
 from .lang.c3 import C3Builder
 from .lang.bf import BrainFuckGenerator
 from .lang.fortran import FortranBuilder
 from .lang.llvmir import LlvmIrFrontend
 from .lang.pascal import PascalBuilder
 from .lang.ws import WhitespaceGenerator
-from .irutils import Verifier
+from .lang.python import python_to_ir, ir_to_python
+from .irs.wasm import wasm_to_ir, read_wasm
+from .irutils import verify_module
 from .utils.reporting import DummyReportGenerator, HtmlReportGenerator
 from .opt.transform import DeleteUnusedInstructionsPass
 from .opt.transform import RemoveAddZeroPass
 from .opt import CommonSubexpressionEliminationPass
 from .opt import ConstantFolder
-from .opt.transform import LoadAfterStorePass
+from .opt import LoadAfterStorePass
 from .opt import CleanPass
 from .opt.mem2reg import Mem2RegPromotor
 from .opt.cjmp import CJumpPass
 from .codegen import CodeGenerator
-from .binutils.linker import Linker
-from .binutils.layout import Layout
+from .binutils.linker import link
 from .binutils.outstream import BinaryOutputStream, TextOutputStream
 from .binutils.outstream import MasterOutputStream, FunctionOutputStream
-from .binutils.objectfile import ObjectFile
-from .binutils.debuginfo import DebugDb, DebugAddress, DebugInfo
+from .binutils.objectfile import ObjectFile, get_object
+from .binutils.debuginfo import DebugAddress, DebugInfo
 from .binutils.disasm import Disassembler
 from .utils.hexfile import HexFile
 from .utils.elffile import ElfFile
 from .utils.exefile import ExeWriter
-from .utils.ir2py import IrToPython
+from .utils import uboot_image
 from .build.tasks import TaskError, TaskRunner
 from .build.recipe import RecipeLoader
-from .common import CompilerError, DiagnosticsManager
+from .common import CompilerError, DiagnosticsManager, get_file
 from .arch.target_list import create_arch
 
 # When using 'from ppci.api import *' include the following:
 __all__ = [
     'asm', 'c3c', 'cc', 'link', 'objcopy', 'bfcompile', 'construct',
-    'optimize',
-    'get_arch', 'ir_to_object']
+    'optimize', 'preprocess',
+    'get_arch', 'ir_to_object', 'ir_to_python']
 
 
 def get_arch(arch):
@@ -96,46 +97,20 @@ def get_reporter(reporter):
         return reporter
 
 
-def get_current_platform():
-    """ Try to get a platform for the current platform """
-    machine = platform.machine()
-    # system = platform.system()
-    arch = get_arch(machine)
-    return arch
+def is_platform_supported():
+    """ Determine if this platform is supported """
+    return get_current_arch() is not None
 
 
-def get_file(f, mode='r'):
-    """ Determine if argument is a file like object or make it so! """
-    if hasattr(f, 'read'):
-        # Assume this is a file like object
-        return f
-    elif isinstance(f, str):
-        try:
-            return open(f, mode)
-        except FileNotFoundError:
-            raise TaskError('Cannot open {}'.format(f))
-    else:
-        raise TaskError('Cannot open {}'.format(f))
-
-
-def get_object(obj):
-    """ Try hard to load an object """
-    if not isinstance(obj, ObjectFile):
-        f = get_file(obj)
-        obj = ObjectFile.load(f)
-        f.close()
-    return obj
-
-
-def get_layout(layout):
-    """ Get a layout from object or file """
-    if isinstance(layout, Layout):
-        return layout
-    else:
-        file = get_file(layout)
-        layout = Layout.load(file)
-        file.close()
-        return layout
+def get_current_arch():
+    """ Try to get the architecture for the current platform """
+    if sys.platform.startswith('win'):
+        machine = platform.machine()
+        if machine == 'AMD64':
+            return get_arch('x86_64:wincc')
+    elif sys.platform == 'linux':
+        if platform.architecture()[0] == '64bit':
+            return get_arch('x86_64')
 
 
 def construct(buildfile, targets=()):
@@ -208,8 +183,9 @@ def asm(source, march, debug=False):
 def disasm(data, march):
     """ Disassemble the given binary data for machine march.
 
-    data can be a filename or a file like object.
-    march can be a machine instance or a string indicating the target.
+    Args:
+        data: a filename or a file like object.
+        march: a machine instance or a string indicating the architecture.
 
     .. doctest::
 
@@ -227,27 +203,24 @@ def disasm(data, march):
     disassembler.disasm(data, ostream)
 
 
-def c3toir(sources, includes, march, reporter=None, debug_db=None):
+def c3toir(sources, includes, march, reporter=None):
     """ Compile c3 sources to ir-code for the given architecture. """
     logger = logging.getLogger('c3c')
     march = get_arch(march)
     if not reporter:  # pragma: no cover
         reporter = DummyReportGenerator()
 
-    if not debug_db:  # pragma: no cover
-        debug_db = DebugDb()
-
     logger.debug('C3 compilation started')
     reporter.heading(2, 'c3 compilation')
     sources = [get_file(fn) for fn in sources]
     includes = [get_file(fn) for fn in includes]
     diag = DiagnosticsManager()
-    c3b = C3Builder(diag, march.info, debug_db)
+    c3b = C3Builder(diag, march.info)
 
     try:
         _, ir_modules = c3b.build(sources, includes)
         for ircode in ir_modules:
-            Verifier().verify(ircode)
+            verify_module(ircode)
     except CompilerError as ex:
         diag.error(ex.msg, ex.loc)
         diag.print_errors()
@@ -264,7 +237,7 @@ def c3toir(sources, includes, march, reporter=None, debug_db=None):
 OPT_LEVELS = ('0', '1', '2', 's')
 
 
-def optimize(ir_module, level=0, reporter=None, debug_db=None):
+def optimize(ir_module, level=0, reporter=None):
     """ Run a bag of tricks against the :doc:`ir-code<ir>`.
 
     This is an in-place operation!
@@ -276,21 +249,17 @@ def optimize(ir_module, level=0, reporter=None, debug_db=None):
             1: some optimization
             2: more optimization
             s: optimize for size
+        reporter: Report detailed log to this reporter
     """
     logger = logging.getLogger('optimize')
     level = str(level)
 
-    if not reporter:  # pragma: no cover
-        reporter = DummyReportGenerator()
-
-    if not debug_db:  # pragma: no cover
-        debug_db = DebugDb()
-
     logger.info('Optimizing module %s level %s', ir_module.name, level)
 
-    reporter.message('{} before optimization:'.format(ir_module))
-    reporter.message('{} {}'.format(ir_module, ir_module.stats()))
-    reporter.dump_ir(ir_module)
+    if reporter:
+        reporter.message('{} before optimization:'.format(ir_module))
+        reporter.message('{} {}'.format(ir_module, ir_module.stats()))
+        reporter.dump_ir(ir_module)
 
     assert level in OPT_LEVELS
     if level == '0':
@@ -298,37 +267,36 @@ def optimize(ir_module, level=0, reporter=None, debug_db=None):
 
     # TODO: differentiate between optimization levels!
 
-    # Create the verifier:
-    verifier = Verifier()
-
     # Optimization passes (bag of tricks) run them three times:
-    opt_passes = [Mem2RegPromotor(debug_db),
-                  RemoveAddZeroPass(debug_db),
-                  ConstantFolder(debug_db),
-                  CommonSubexpressionEliminationPass(debug_db),
-                  LoadAfterStorePass(debug_db),
-                  DeleteUnusedInstructionsPass(debug_db),
-                  CleanPass(debug_db)] * 3
+    opt_passes = [Mem2RegPromotor(),
+                  RemoveAddZeroPass(),
+                  ConstantFolder(),
+                  CommonSubexpressionEliminationPass(),
+                  LoadAfterStorePass(),
+                  DeleteUnusedInstructionsPass(),
+                  CleanPass()] * 3
 
     if level == '3':
-        opt_passes.append(CJumpPass(debug_db))
+        opt_passes.append(CJumpPass())
 
     # Run the passes over the module:
-    verifier.verify(ir_module)
+    verify_module(ir_module)
     for opt_pass in opt_passes:
         opt_pass.run(ir_module)
         # reporter.message('{} after {}:'.format(ir_module, opt_pass))
         # reporter.dump_ir(ir_module)
-    verifier.verify(ir_module)
 
-    # Dump report:
-    reporter.message('{} after optimization:'.format(ir_module))
-    reporter.message('{} {}'.format(ir_module, ir_module.stats()))
-    reporter.dump_ir(ir_module)
+    if reporter:
+        # Dump report:
+        reporter.message('{} after optimization:'.format(ir_module))
+        reporter.message('{} {}'.format(ir_module, ir_module.stats()))
+        reporter.dump_ir(ir_module)
+
+    verify_module(ir_module)
 
 
 def ir_to_stream(
-        ir_module, march, output_stream, debug_db=None, reporter=None,
+        ir_module, march, output_stream, reporter=None,
         debug=False, opt='speed'):
     """ Translate IR module to output stream.
     """
@@ -337,12 +305,8 @@ def ir_to_stream(
     if not reporter:  # pragma: no cover
         reporter = DummyReportGenerator()
 
-    if not debug_db:  # pragma: no cover
-        debug_db = DebugDb()
-
-    code_generator = CodeGenerator(march, debug_db, optimize_for=opt)
-    verifier = Verifier()
-    verifier.verify(ir_module)
+    code_generator = CodeGenerator(march, optimize_for=opt)
+    verify_module(ir_module)
 
     # Code generation:
     code_generator.generate(
@@ -350,7 +314,7 @@ def ir_to_stream(
 
 
 def ir_to_object(
-        ir_modules, march, debug_db=None, reporter=None, debug=False,
+        ir_modules, march, reporter=None, debug=False,
         opt='speed', outstream=None):
     """ Translate IR-modules into code for the given architecture.
 
@@ -371,9 +335,6 @@ def ir_to_object(
     if not reporter:  # pragma: no cover
         reporter = DummyReportGenerator()
 
-    if not debug_db:  # pragma: no cover
-        debug_db = DebugDb()
-
     reporter.heading(2, 'Code generation')
     reporter.message("Target: {}".format(march))
 
@@ -394,7 +355,7 @@ def ir_to_object(
     for ir_module in ir_modules:
         ir_to_stream(
             ir_module, march, output_stream,
-            debug_db=debug_db, reporter=reporter, debug=debug, opt=opt)
+            reporter=reporter, debug=debug, opt=opt)
 
     # TODO: refactor polishing?
     obj.polish()
@@ -403,52 +364,7 @@ def ir_to_object(
     return obj
 
 
-def ir_to_python(ir_modules, f, reporter=None):
-    """ Convert ir-code to python code """
-    if not reporter:  # pragma: no cover
-        reporter = DummyReportGenerator()
-    generator = IrToPython(f)
-    generator.header()
-    for ir_module in ir_modules:
-        generator.generate(ir_module)
-
-
-def preprocess(f, output_file, coptions=None):
-    """ Pre-process a file into the other file. """
-    if coptions is None:
-        coptions = COptions()
-    preprocessor = CPreProcessor(coptions)
-    filename = f.name if hasattr(f, 'name') else None
-    tokens = preprocessor.process(f, filename=filename)
-    CTokenPrinter().dump(tokens, file=output_file)
-
-
-def c_to_ir(
-        source: io.TextIOBase, march, coptions=None, debug_db=None,
-        reporter=None):
-    """ C to ir translation. """
-    if not reporter:  # pragma: no cover
-        reporter = DummyReportGenerator()
-
-    if not coptions:  # pragma: no cover
-        coptions = COptions()
-
-    if not debug_db:  # pragma: no cover
-        debug_db = DebugDb()
-
-    march = get_arch(march)
-    cbuilder = CBuilder(march.info, coptions)
-    assert isinstance(source, io.TextIOBase)
-    if hasattr(source, 'name'):
-        filename = getattr(source, 'name')
-    else:
-        filename = None
-    ir_module = cbuilder.build(
-        source, filename, debug_db=debug_db, reporter=reporter)
-    return ir_module
-
-
-def cc(source: io.TextIOBase, march, coptions=None,
+def cc(source: io.TextIOBase, march, coptions=None, opt_level=0,
        debug=False, reporter=None):
     """ C compiler. compiles a single source file into an object file.
 
@@ -476,17 +392,25 @@ def cc(source: io.TextIOBase, march, coptions=None,
 
     if not coptions:
         coptions = COptions()
-    debug_db = DebugDb()
 
-    ir_module = c_to_ir(
-        source, march, coptions=coptions, debug_db=debug_db,
-        reporter=reporter)
+    ir_module = c_to_ir(source, march, coptions=coptions, reporter=reporter)
     reporter.message('{} {}'.format(ir_module, ir_module.stats()))
     reporter.dump_ir(ir_module)
-    return ir_to_object(
-        [ir_module], march,
-        debug_db=debug_db, debug=debug,
-        reporter=reporter)
+    optimize(ir_module, level=opt_level, reporter=reporter)
+    return ir_to_object([ir_module], march, debug=debug, reporter=reporter)
+
+
+def wasmcompile(source: io.TextIOBase, march, opt_level=2):
+    """ Webassembly compile """
+    march = get_arch(march)
+    wasm_module = read_wasm(source)
+    ir_module = wasm_to_ir(wasm_module)
+
+    # Optimize:
+    optimize(ir_module, level=opt_level)
+
+    obj = ir_to_object([ir_module], march)
+    return obj
 
 
 def llvm_to_ir(source):
@@ -530,17 +454,15 @@ def c3c(sources, includes, march, opt_level=0, reporter=None, debug=False,
     """
     reporter = get_reporter(reporter)
     march = get_arch(march)
-    debug_db = DebugDb()
     ir_modules = \
-        c3toir(sources, includes, march, reporter=reporter, debug_db=debug_db)
+        c3toir(sources, includes, march, reporter=reporter)
 
     for ircode in ir_modules:
         optimize(ircode, level=opt_level, reporter=reporter)
 
     opt_cg = 'size' if opt_level == 's' else 'speed'
     return ir_to_object(
-        ir_modules, march,
-        debug_db=debug_db, debug=debug, reporter=reporter,
+        ir_modules, march, debug=debug, reporter=reporter,
         opt=opt_cg, outstream=outstream)
 
 
@@ -558,8 +480,7 @@ def pascal(sources, march, opt_level=0, reporter=None):
     march = get_arch(march)
     if not reporter:  # pragma: no cover
         reporter = DummyReportGenerator()
-    debug_db = DebugDb()
-    pascal_builder = PascalBuilder(diag, march.info, debug_db)
+    pascal_builder = PascalBuilder(diag, march.info)
 
     sources = [get_file(fn) for fn in sources]
     ir_modules = pascal_builder.build(sources)
@@ -614,8 +535,18 @@ def bfcompile(source, target, reporter=None):
         'Before optimization {} {}'.format(ir_module, ir_module.stats()))
     reporter.dump_ir(ir_module)
     optimize(ir_module, reporter=reporter)
-    debug_db = DebugDb()
-    return ir_to_object([ir_module], target, debug_db, reporter=reporter)
+    return ir_to_object([ir_module], target, reporter=reporter)
+
+
+def pycompile(source, march, reporter=None):
+    """ Compile a piece of python code to machine code.
+
+    Note that the python code must be type annotated for this
+    to work.
+    """
+    march = get_arch(march)
+    ir_module = python_to_ir(source)
+    return ir_to_object([ir_module], march)
 
 
 def fortrancompile(sources, target, reporter=DummyReportGenerator()):
@@ -630,58 +561,9 @@ def llvmir2ir(f):
     return LlvmIrFrontend().compile(f)
 
 
-def link(
-        objects, layout=None, use_runtime=False, partial_link=False,
-        reporter=None, debug=False):
-    """ Links the iterable of objects into one using the given layout.
-
-    Args:
-        objects: a collection of objects to be linked together.
-        use_runtime (bool): also link compiler runtime functions
-        debug (bool): when true, keep debug information. Otherwise remove
-            this debug information from the result.
-
-    Returns:
-        The linked object file
-
-    .. doctest::
-
-        >>> import io
-        >>> from ppci.api import asm, c3c, link
-        >>> asm_source = io.StringIO("db 0x77")
-        >>> obj1 = asm(asm_source, 'arm')
-        >>> c3_source = io.StringIO("module main; var int a;")
-        >>> obj2 = c3c([c3_source], [], 'arm')
-        >>> obj = link([obj1, obj2])
-        >>> print(obj)
-        CodeObject of 8 bytes
-    """
-    if not reporter:  # pragma: no cover
-        reporter = DummyReportGenerator()
-
-    objects = [get_object(obj) for obj in objects]
-    if not objects:
-        raise ValueError('Please provide at least one object as input')
-
-    if layout:
-        layout = get_layout(layout)
-
-    march = objects[0].arch
-
-    if use_runtime:
-        objects.append(march.runtime)
-
-    linker = Linker(march, reporter)
-    try:
-        output_obj = linker.link(objects, layout=layout, debug=debug)
-    except CompilerError as err:
-        raise TaskError(err.msg)
-    return output_obj
-
-
 def objcopy(obj: ObjectFile, image_name: str, fmt: str, output_filename):
     """ Copy some parts of an object file to an output """
-    fmts = ['bin', 'hex', 'elf', 'exe', 'ldb']
+    fmts = ['bin', 'hex', 'elf', 'exe', 'ldb', 'uimage']
     if fmt not in fmts:
         formats = ', '.join(fmts[:-1]) + ' and ' + fmts[-1]
         raise TaskError('Only {} are supported'.format(formats))
@@ -700,13 +582,25 @@ def objcopy(obj: ObjectFile, image_name: str, fmt: str, output_filename):
     elif fmt == "hex":
         image = obj.get_image(image_name)
         hexfile = HexFile()
-        hexfile.add_region(image.location, image.data)
+        hexfile.add_region(image.address, image.data)
         with open(output_filename, 'w') as output_file:
             hexfile.save(output_file)
     elif fmt == 'ldb':
         # TODO: fix this some other way to extract debug info
         with open(output_filename, 'w') as output_file:
             write_ldb(obj, output_file)
+    elif fmt == 'uimage':
+        image = obj.get_image(image_name)
+        uboot_architectures = {
+            'arm': uboot_image.Architecture.ARM,
+            'or1k': uboot_image.Architecture.OPENRISC,
+            'xtensa': uboot_image.Architecture.XTENSA,
+        }
+        with open(output_filename, 'wb') as f:
+            uboot_image.write_uboot_image(
+                f, image.data,
+                load_address=image.address, entry_point=image.address,
+                arch=uboot_architectures[obj.arch.name])
     elif fmt == 'exe':
         writer = ExeWriter()
         with open(output_filename, 'wb') as output_file:
