@@ -201,7 +201,7 @@ class regregimm3_base(ThumbInstruction):
         return tokens[0].encode()
 
 
-class Add2(regregimm3_base):
+class AddImm(regregimm3_base):
     """ add Rd, Rn, imm3 """
     syntax = Syntax([
         'add', ' ', regregimm3_base.rd, ',', ' ', regregimm3_base.rn, ',',
@@ -209,7 +209,7 @@ class Add2(regregimm3_base):
     opcode = 0b0001110
 
 
-class Sub2(regregimm3_base):
+class SubImm(regregimm3_base):
     """ sub Rd, Rn, imm3 """
     syntax = Syntax([
         'sub', ' ', regregimm3_base.rd, ',', ' ', regregimm3_base.rn, ',',
@@ -333,6 +333,7 @@ Eor = make_regreg('eor', 0b0100000001)
 Cmp = make_regreg('cmp', 0b0100001010)
 Lsl = make_regreg('lsl', 0b0100000010)
 Lsr = make_regreg('lsr', 0b0100000011)
+Rsb = make_regreg('rsb', 0b0100001001)
 
 
 class Cmp2(ThumbInstruction):
@@ -409,6 +410,19 @@ class Bl(LongThumbInstruction):
         return [BlImm11Relocation(self.target)]
 
 
+class Blx(ThumbInstruction):
+    """ Branch with link with target in a register """
+    rm = Operand('rm', ArmRegister, read=True)
+    syntax = Syntax(['blx', ' ', rm])
+
+    def encode(self):
+        tokens = self.get_tokens()
+        tokens[0][7:16] = 0b010001111
+        tokens[0][3:7] = self.rm.num
+        tokens[0][0:3] = 0
+        return tokens.encode()
+
+
 class cond_base_ins(ThumbInstruction):
     def encode(self):
         imm8 = 0
@@ -576,12 +590,6 @@ class SubSp(addspsp_base):
 #
 ###############
 
-@thumb_isa.pattern('stm', 'STRI32(reg, reg)', size=2)
-@thumb_isa.pattern('stm', 'STRU32(reg, reg)', size=2)
-def pattern_str32(context, tree, c0, c1):
-    context.emit(Str2(c1, c0, 0))
-
-
 @thumb_isa.pattern('stm', 'JMP', size=2)
 def pattern_jmp(context, tree):
     label = tree.value
@@ -644,10 +652,10 @@ def pattern_label(context, tree):
     'reg', 'FPRELU32', size=9, cycles=9)
 def pattern_fprel32(context, tree):
     d = context.new_reg(LowArmRegister)
-    c1 = tree.value.negative
+    c1 = tree.value.offset
     if c1 >= 0:
         if c1 < 8:
-            context.emit(Add2(d, R7, c1))
+            context.emit(AddImm(d, R7, c1))
         elif c1 < 256:
             d2 = context.new_reg(LowArmRegister)
             context.emit(Mov3(d2, c1))
@@ -660,7 +668,7 @@ def pattern_fprel32(context, tree):
     else:
         c1 = -c1
         if c1 < 8:
-            context.emit(Sub2(d, R7, c1))
+            context.emit(SubImm(d, R7, c1))
         elif c1 < 256:
             d2 = context.new_reg(LowArmRegister)
             context.emit(Mov3(d2, c1))
@@ -745,28 +753,44 @@ def pattern_cjmp(context, tree, c0, c1):
     context.emit(jmp_ins)
 
 
-@thumb_isa.pattern('stm', 'STRI8(reg,reg)', size=2)
-@thumb_isa.pattern('stm', 'STRU8(reg,reg)', size=2)
+@thumb_isa.pattern('mem', 'reg', size=0)
+def pattern_mem_reg(context, tree, c0):
+    return c0, 0
+
+
+@thumb_isa.pattern('stm', 'STRI8(mem,reg)', size=2)
+@thumb_isa.pattern('stm', 'STRU8(mem,reg)', size=2)
 def pattern_str8(context, tree, c0, c1):
-    context.emit(Strb(c1, c0, 0))
+    rg, offset = c0
+    context.emit(Strb(c1, rg, offset))
 
 
-@thumb_isa.pattern('reg', 'LDRI8(reg)', size=2)
-@thumb_isa.pattern('reg', 'LDRU8(reg)', size=2)
+@thumb_isa.pattern('stm', 'STRI32(mem, reg)', size=2)
+@thumb_isa.pattern('stm', 'STRU32(mem, reg)', size=2)
+def pattern_str32(context, tree, c0, c1):
+    rg, offset = c0
+    context.emit(Str2(c1, rg, offset))
+
+
+@thumb_isa.pattern('reg', 'LDRI8(mem)', size=2)
+@thumb_isa.pattern('reg', 'LDRU8(mem)', size=2)
 def pattern_ldr8(context, tree, c0):
+    rg, offset = c0
     d = context.new_reg(LowArmRegister)
-    context.emit(Ldrb(d, c0, 0))
+    context.emit(Ldrb(d, rg, offset))
     return d
 
 
-@thumb_isa.pattern('reg', 'LDRI32(reg)', size=2)
-@thumb_isa.pattern('reg', 'LDRU32(reg)', size=2)
+@thumb_isa.pattern('reg', 'LDRI32(mem)', size=2)
+@thumb_isa.pattern('reg', 'LDRU32(mem)', size=2)
 def pattern_ldr32(context, tree, c0):
+    rg, offset = c0
     d = context.new_reg(LowArmRegister)
-    context.emit(Ldr2(d, c0, 0))
+    context.emit(Ldr2(d, rg, offset))
     return d
 
 
+# Arithmatic patterns:
 @thumb_isa.pattern('reg', 'SUBU32(reg,reg)', size=2)
 @thumb_isa.pattern('reg', 'SUBI32(reg,reg)', size=2)
 def pattern_sub32(context, tree, c0, c1):
@@ -777,11 +801,11 @@ def pattern_sub32(context, tree, c0, c1):
 
 @thumb_isa.pattern(
     'reg', 'SUBI32(reg,CONSTI32)', size=2,
-    condition=lambda x: x.children[1].value in range(8))
+    condition=lambda x: x[1].value in range(8))
 def pattern_sub32_imm3(context, tree, c0):
     d = context.new_reg(LowArmRegister)
-    c1 = tree.children[1].value
-    context.emit(Sub2(d, c0, c1))
+    c1 = tree[1].value
+    context.emit(SubImm(d, c0, c1))
     return d
 
 
@@ -801,6 +825,14 @@ def pattern_shr32(context, tree, c0, c1):
     return d
 
 
+@thumb_isa.pattern('reg', 'NEGI32(reg)', size=2)
+@thumb_isa.pattern('reg', 'NEGU32(reg)', size=2)
+def pattern_neg32(context, tree, c0):
+    d = context.new_reg(LowArmRegister)
+    context.emit(Rsb(d, c0))
+    return d
+
+
 @thumb_isa.pattern('reg', 'ORI32(reg, reg)', size=4)
 @thumb_isa.pattern('reg', 'ORU32(reg, reg)', size=4)
 def pattern_or32(context, tree, c0, c1):
@@ -829,6 +861,7 @@ def pattern_shl32(context, tree, c0, c1):
 
 
 @thumb_isa.pattern('reg', 'MULI32(reg, reg)', size=5)
+@thumb_isa.pattern('reg', 'MULU32(reg, reg)', size=5)
 def pattern_mul32(context, tree, c0, c1):
     d = context.new_reg(LowArmRegister)
     context.move(d, c0)
