@@ -4,19 +4,17 @@
 
 from ..isa import Isa
 from ..encoding import Instruction, Syntax, Operand
-from ..data_instructions import Dd, DataInstruction, ByteToken, WordToken
-from ..data_instructions import DByte, DZero
-from ...utils.bitfun import wrap_negative, inrange
+from ..data_instructions import Dd
+from ...utils.bitfun import inrange
 from ..generic_instructions import ArtificialInstruction, Alignment
 from ..generic_instructions import SectionInstruction
-from .registers import RiscvRegister, FP, LR, R0, R10, R12, R13
+from .registers import RiscvRegister, FP, R0
 from .relocations import AbsAddr32Relocation
 from .relocations import BImm12Relocation, BImm20Relocation
 from .relocations import Abs32Imm20Relocation
 from .relocations import Abs32Imm12Relocation, RelImm20Relocation
 from .relocations import RelImm12Relocation
 from .tokens import RiscvToken, RiscvIToken
-from ...ir import i32
 
 isa = Isa()
 
@@ -145,12 +143,13 @@ class IBase(RiscvInstruction):
         tokens[0][7:12] = self.rd.num
         tokens[0][12:15] = self.func
         tokens[0][15:20] = self.rs1.num
-        self.imm = wrap_negative(self.imm, 12)
+        self.imm = self.imm & 0xfff
         tokens[0][20:32] = self.imm
         return tokens[0].encode()
 
 
 def make_i(mnemonic, func):
+    """ Factory function for immediate value instructions """
     rd = Operand('rd', RiscvRegister, write=True)
     rs1 = Operand('rs1', RiscvRegister, read=True)
     imm = Operand('imm', int)
@@ -263,7 +262,7 @@ class Lui(RiscvInstruction):
 
     def encode(self):
         tokens = self.get_tokens()
-        imm20 = wrap_negative(self.imm, 20)
+        imm20 = self.imm & 0xfffff
         tokens[0][0:7] = 0b0110111
         tokens[0][7:12] = self.rd.num
         tokens[0][12:32] = imm20
@@ -321,7 +320,7 @@ class Adrl(RiscvInstruction):
         return [Abs32Imm12Relocation(self.label)]
 
 
-class Adrlrel(RiscvInstruction):
+class Loadlrel(RiscvInstruction):
     rd = Operand('rd', RiscvRegister, write=True, read=True)
     label = Operand('label', str)
     syntax = Syntax(['lw', ' ', rd, ',', ' ', label])
@@ -339,6 +338,24 @@ class Adrlrel(RiscvInstruction):
         return [RelImm12Relocation(self.label)]
 
 
+class Adrlrel(RiscvInstruction):
+    rd = Operand('rd', RiscvRegister, write=True, read=True)
+    label = Operand('label', str)
+    syntax = Syntax(['addi', ' ', rd, ',', ' ', label])
+
+    def encode(self):
+        tokens = self.get_tokens()
+        tokens[0][0:7] = 0b0010011
+        tokens[0][7:12] = self.rd.num
+        tokens[0][12:15] = 0b000
+        tokens[0][15:20] = self.rd.num
+        tokens[0][20:32] = 0
+        return tokens[0].encode()
+
+    def relocations(self):
+        return [RelImm12Relocation(self.label)]
+
+
 class Auipc(RiscvInstruction):
     rd = Operand('rd', RiscvRegister, write=True)
     imm = Operand('imm', int)
@@ -352,6 +369,16 @@ class Auipc(RiscvInstruction):
         return tokens[0].encode()
 
 
+class Labelrel(PseudoRiscvInstruction):
+    rd = Operand('rd', RiscvRegister, write=True)
+    label = Operand('label', str)
+    syntax = Syntax(['lw', ' ', rd, ',', ' ', label])
+
+    def render(self):
+        yield Adrurel(self.rd, self.label)
+        yield Loadlrel(self.rd, self.label)
+
+
 class La(PseudoRiscvInstruction):
     rd = Operand('rd', RiscvRegister, write=True)
     label = Operand('label', str)
@@ -368,14 +395,15 @@ class Li(PseudoRiscvInstruction):
     syntax = Syntax(['li', ' ', rd, ',', ' ', imm])
 
     def render(self):
-        if (inrange(self.imm, 12)) is False:
+        # If the immediate value fits into 12 bits, do so!
+        if inrange(self.imm, 12):
+            yield Addi(self.rd, R0, self.imm)
+        else:
             if (self.imm & 0x800) != 0:
                 self.imm += 0x1000
             yield Lui(self.rd, self.imm >> 12)
-            self.imm = wrap_negative(self.imm, 12)
-            yield Addi(self.rd, self.rd, self.imm)
-        else:
-            yield Addi(self.rd, R0, self.imm)
+            lower_bits = self.imm & 0xfff
+            yield Addi(self.rd, self.rd, lower_bits)
 
 
 class BranchBase(RiscvInstruction):
@@ -412,10 +440,13 @@ def make_branch(mnemonic, cond, invert):
 Beq = make_branch('beq', 0b000, False)
 Bne = make_branch('bne', 0b001, False)
 Blt = make_branch('blt', 0b100, False)
-Bltu = make_branch('bltu', 0b110, False)
-Bge = make_branch('bge', 0b101, False)
-Bgeu = make_branch('bgeu', 0b111, False)
 Bgt = make_branch('bgt', 0b100, True)
+Bge = make_branch('bge', 0b101, False)
+Ble = make_branch('bge', 0b101, True)
+Bltu = make_branch('bltu', 0b110, False)
+Bgtu = make_branch('bgtu', 0b110, True)
+Bgeu = make_branch('bgeu', 0b111, False)
+Bleu = make_branch('bleu', 0b111, True)
 
 
 def reg_list_to_mask(reg_list):
@@ -427,11 +458,8 @@ def reg_list_to_mask(reg_list):
 
 class StrBase(RiscvInstruction):
     def encode(self):
-        if self.offset < 0:
-            imml5 = wrap_negative(-((-self.offset) & 0x1f), 5)
-        else:
-            imml5 = self.offset & 0x1f
-        immh7 = wrap_negative(self.offset >> 5, 7)
+        imml5 = self.offset & 0x1f
+        immh7 = (self.offset >> 5) & 0x7f
         tokens = self.get_tokens()
         tokens[0][0:7] = 0b0100011
         tokens[0][7:12] = imml5
@@ -507,45 +535,14 @@ def make_mext(mnemonic, func):
 
 Mul = make_mext('mul', 0b000)
 Div = make_mext('div', 0b100)
+Divu = make_mext('divu', 0b101)
+Rem = make_mext('rem', 0b110)
+Remu = make_mext('remu', 0b111)
 
 
 # Instruction selection patterns:
-@isa.pattern('stm', 'STRU32(reg, reg)', size=2)
-@isa.pattern('stm', 'STRI32(reg, reg)', size=2)
-def pattern_str_i32(self, tree, c0, c1):
-    self.emit(Sw(c1, 0, c0))
 
 
-@isa.pattern(
-    'stm', 'STRI32(ADDI32(reg, CONSTI32), reg)',
-    size=2,
-    condition=lambda t: t.children[0].children[1].value < 2047)
-def pattern_str_i32_add(context, tree, c0, c1):
-    # TODO: something strange here: when enabeling this rule, programs
-    # compile correctly...
-    offset = tree.children[0].children[1].value
-    context.emit(Sw(c1, offset, c0))
-
-@isa.pattern(
-    'stm', 'STRI32(ADDU32(reg, CONSTU32), reg)',
-    size=2,
-    condition=lambda t: t.children[0].children[1].value < 2047)
-def pattern_str_i32_add(context, tree, c0, c1):
-    # TODO: something strange here: when enabeling this rule, programs
-    # compile correctly...
-    offset = tree.children[0].children[1].value
-    context.emit(Sw(c1, offset, c0))
-
-@isa.pattern('stm', 'STRI16(reg, reg)', size=2)
-@isa.pattern('stm', 'STRU16(reg, reg)', size=2)
-def pattern_str16(context, tree, c0, c1):
-    context.emit(Sh(c1, 0, c0))
-
-@isa.pattern('stm', 'STRI8(reg, reg)', size=2)
-@isa.pattern('stm', 'STRU8(reg, reg)', size=2)
-def pattern_str8(context, tree, c0, c1):
-    context.emit(Sb(c1, 0, c0))
-
 @isa.pattern('reg', 'MOVI16(reg)', size=2)
 @isa.pattern('reg', 'MOVU16(reg)', size=2)
 @isa.pattern('reg', 'MOVI32(reg)', size=2)
@@ -577,25 +574,55 @@ def pattern_jmp(context, tree):
 def pattern_reg(context, tree):
     return tree.value
 
-@isa.pattern('reg', 'I32TOI32(reg)', size=0)
-@isa.pattern('reg', 'I32TOU32(reg)', size=0)
-@isa.pattern('reg', 'U32TOI32(reg)', size=0)
+
 @isa.pattern('reg', 'U32TOU32(reg)', size=0)
+@isa.pattern('reg', 'U32TOI32(reg)', size=0)
+@isa.pattern('reg', 'I32TOU32(reg)', size=0)
+@isa.pattern('reg', 'I32TOI32(reg)', size=0)
+@isa.pattern('reg', 'U32TOU16(reg)', size=0)
+@isa.pattern('reg', 'U32TOI16(reg)', size=0)
+@isa.pattern('reg', 'I32TOI16(reg)', size=0)
+@isa.pattern('reg', 'I32TOU16(reg)', size=0)
+@isa.pattern('reg', 'U16TOU16(reg)', size=0)
+@isa.pattern('reg', 'U16TOI16(reg)', size=0)
+@isa.pattern('reg', 'I16TOI16(reg)', size=0)
+@isa.pattern('reg', 'I16TOU16(reg)', size=0)
+@isa.pattern('reg', 'U16TOU8(reg)', size=0)
+@isa.pattern('reg', 'U16TOI8(reg)', size=0)
+@isa.pattern('reg', 'I16TOI8(reg)', size=0)
+@isa.pattern('reg', 'I16TOU8(reg)', size=0)
+@isa.pattern('reg', 'U8TOU8(reg)', size=0)
+@isa.pattern('reg', 'U8TOI8(reg)', size=0)
+@isa.pattern('reg', 'I8TOI8(reg)', size=0)
+@isa.pattern('reg', 'I8TOU8(reg)', size=0)
 def pattern_i32_to_i32(context, tree, c0):
     return c0
 
+
+@isa.pattern('reg', 'I8TOI16(reg)', size=4)
 @isa.pattern('reg', 'I8TOI32(reg)', size=4)
 def pattern_i8_to_i32(context, tree, c0):
     context.emit(Slli(c0, c0, 24))
     context.emit(Srai(c0, c0, 24))
     return c0
 
+
 @isa.pattern('reg', 'I16TOI32(reg)', size=4)
-def pattern_i8_to_i32(context, tree, c0):
+def pattern_i16_to_i32(context, tree, c0):
     context.emit(Slli(c0, c0, 16))
     context.emit(Srai(c0, c0, 16))
     return c0
 
+
+@isa.pattern('reg', 'I8TOU16(reg)', size=4)
+@isa.pattern('reg', 'U8TOU16(reg)', size=4)
+@isa.pattern('reg', 'U8TOI16(reg)', size=4)
+def pattern_8_to_16(context, tree, c0):
+    context.emit(Slli(c0, c0, 24))
+    context.emit(Srli(c0, c0, 24))
+    return c0
+
+
 @isa.pattern('reg', 'I8TOU32(reg)', size=4)
 @isa.pattern('reg', 'U8TOU32(reg)', size=4)
 @isa.pattern('reg', 'U8TOI32(reg)', size=4)
@@ -604,6 +631,7 @@ def pattern_8_to_32(context, tree, c0):
     context.emit(Srli(c0, c0, 24))
     return c0
 
+
 @isa.pattern('reg', 'I16TOU32(reg)', size=4)
 @isa.pattern('reg', 'U16TOU32(reg)', size=4)
 @isa.pattern('reg', 'U16TOI32(reg)', size=4)
@@ -612,6 +640,7 @@ def pattern_16_to_32(context, tree, c0):
     context.emit(Srli(c0, c0, 16))
     return c0
 
+
 @isa.pattern('reg', 'I32TOI8(reg)', size=0)
 @isa.pattern('reg', 'I32TOU8(reg)', size=0)
 @isa.pattern('reg', 'I32TOI16(reg)', size=0)
@@ -625,10 +654,10 @@ def pattern_32_to_8_16(context, tree, c0):
     return c0
 
 
-
 @isa.pattern('reg', 'CONSTI32', size=4)
 @isa.pattern('reg', 'CONSTU32', size=4)
 @isa.pattern('reg', 'CONSTI16', size=4)
+@isa.pattern('reg', 'CONSTU16', size=4)
 @isa.pattern(
     'reg', 'CONSTI32', size=2,
     condition=lambda t: t.value in range(-2048, 2048))
@@ -647,19 +676,30 @@ def pattern_const_i32(context, tree):
 
 
 @isa.pattern('stm', 'CJMPI32(reg, reg)', size=2)
+@isa.pattern('stm', 'CJMPI16(reg, reg)', size=2)
 @isa.pattern('stm', 'CJMPI8(reg, reg)', size=2)
 def pattern_cjmp(context, tree, c0, c1):
     op, yes_label, no_label = tree.value
-    opnames = {"<": Blt, ">": Bgt, "==": Beq, "!=": Bne, ">=": Bge, "<=": Bgt}
+    opnames = {"<": Blt, ">": Bgt, "==": Beq, "!=": Bne, ">=": Bge, "<=": Ble}
     Bop = opnames[op]
-    if op == "<=":
-        jmp_ins = B(yes_label.name, jumps=[yes_label])
-        context.emit(Bop(c0, c1, yes_label.name, jumps=[no_label, jmp_ins]))
-        context.emit(jmp_ins)
-    else:
-        jmp_ins = B(no_label.name, jumps=[no_label])
-        context.emit(Bop(c0, c1, yes_label.name, jumps=[yes_label, jmp_ins]))
-        context.emit(jmp_ins)
+    jmp_ins = B(no_label.name, jumps=[no_label])
+    context.emit(Bop(c0, c1, yes_label.name, jumps=[yes_label, jmp_ins]))
+    context.emit(jmp_ins)
+
+
+@isa.pattern('stm', 'CJMPU8(reg, reg)', size=2)
+@isa.pattern('stm', 'CJMPU16(reg, reg)', size=2)
+@isa.pattern('stm', 'CJMPU32(reg, reg)', size=2)
+def pattern_cjmpu(context, tree, c0, c1):
+    op, yes_label, no_label = tree.value
+    opnames = {
+        "<": Bltu, ">": Bgtu, "==": Beq, "!=": Bne, ">=": Bgeu, "<=": Bleu
+    }
+    Bop = opnames[op]
+    jmp_ins = B(no_label.name, jumps=[no_label])
+    context.emit(Bop(c0, c1, yes_label.name, jumps=[yes_label, jmp_ins]))
+    context.emit(jmp_ins)
+
 
 @isa.pattern('reg', 'ADDU32(reg, reg)', size=2)
 @isa.pattern('reg', 'ADDI32(reg, reg)', size=2)
@@ -669,6 +709,14 @@ def pattern_add_i32(context, tree, c0, c1):
     return d
 
 
+@isa.pattern('reg', 'ADDU16(reg, reg)', size=2)
+@isa.pattern('reg', 'ADDI16(reg, reg)', size=2)
+def pattern_add_i16(context, tree, c0, c1):
+    d = context.new_reg(RiscvRegister)
+    context.emit(Addr(d, c0, c1))
+    return d
+
+
 @isa.pattern('reg', 'ADDI8(reg, reg)', size=2)
 @isa.pattern('reg', 'ADDU8(reg, reg)', size=2)
 def pattern_add8(context, tree, c0, c1):
@@ -679,10 +727,10 @@ def pattern_add8(context, tree, c0, c1):
 
 @isa.pattern(
     'reg', 'ADDI32(reg, CONSTI32)', size=2,
-    condition=lambda t: t.children[1].value < 2048)
+    condition=lambda t: t[1].value < 2048)
 @isa.pattern(
     'reg', 'ADDU32(reg, CONSTU32)', size=2,
-    condition=lambda t: t.children[1].value < 2048)
+    condition=lambda t: t[1].value < 2048)
 def pattern_add_i32_reg_const(context, tree, c0):
     d = context.new_reg(RiscvRegister)
     c1 = tree.children[1].value
@@ -702,6 +750,9 @@ def pattern_add_i32_const_reg(context, tree, c0):
     context.emit(Addi(d, c0, c1))
     return d
 
+
+@isa.pattern('reg', 'SUBI8(reg, reg)', size=2)
+@isa.pattern('reg', 'SUBU8(reg, reg)', size=2)
 @isa.pattern('reg', 'SUBI16(reg, reg)', size=2)
 @isa.pattern('reg', 'SUBU16(reg, reg)', size=2)
 @isa.pattern('reg', 'SUBI32(reg, reg)', size=2)
@@ -712,14 +763,6 @@ def pattern_sub_i32(context, tree, c0, c1):
     return d
 
 
-@isa.pattern('reg', 'SUBI8(reg, reg)', size=2)
-def pattern_sub_i8(context, tree, c0, c1):
-    # TODO: temporary fix this with an 32 bits sub
-    d = context.new_reg(RiscvRegister)
-    context.emit(Subr(d, c0, c1))
-    return d
-
-
 @isa.pattern('reg', 'LABEL', size=6)
 def pattern_label1(context, tree):
     d = context.new_reg(RiscvRegister)
@@ -734,44 +777,122 @@ def pattern_label1(context, tree):
 def pattern_label2(context, tree):
     d = context.new_reg(RiscvRegister)
     ln = context.frame.add_constant(tree.value)
-    context.emit(La(d, ln))
+    context.emit(Labelrel(d, ln))
     return d
 
 
 @isa.pattern(
-    'reg', 'FPRELU32', size=8,
+    'reg', 'FPRELU32', size=4,
     condition=lambda t: t.value.offset in range(-2048, 2048))
 def pattern_fpreli32(context, tree):
     d = context.new_reg(RiscvRegister)
-    offset = tree.value.negative
+    offset = tree.value.offset
     context.emit(Addi(d, FP, offset))
     return d
 
 
-@isa.pattern('reg', 'LDRU8(reg)', size=2)
-def pattern_ldru8(context, tree, c0):
-    d = context.new_reg(RiscvRegister)
-    context.emit(Lbu(d, 0, c0))
-    return d
+# Memory patterns:
+@isa.pattern(
+    'mem', 'FPRELU32', size=0,
+    condition=lambda t: t.value.offset in range(-2048, 2048))
+def pattern_mem_fpreli32(context, tree):
+    offset = tree.value.offset
+    return FP, offset
 
-@isa.pattern('reg', 'LDRI8(reg)', size=2)
+
+@isa.pattern('mem', 'reg', size=0)
+def pattern_mem_reg(context, tree, c0):
+    return c0, 0
+
+
+@isa.pattern(
+    'mem', 'ADDI32(reg, CONSTI32)', size=0,
+    condition=lambda t: t[1].value < 2047)
+@isa.pattern(
+    'mem', 'ADDU32(reg, CONSTU32)', size=0,
+    condition=lambda t: t[1].value < 2047)
+def pattern_mem_add_reg_const(context, tree, c0):
+    offset = tree[1].value
+    return c0, offset
+
+
+@isa.pattern('stm', 'STRU32(mem, reg)', size=2)
+@isa.pattern('stm', 'STRI32(mem, reg)', size=2)
+def pattern_sw32(context, tree, c0, c1):
+    base_reg, offset = c0
+    context.emit(Sw(c1, offset, base_reg))
+
+
+@isa.pattern('stm', 'STRI16(mem, reg)', size=2)
+@isa.pattern('stm', 'STRU16(mem, reg)', size=2)
+def pattern_str16(context, tree, c0, c1):
+    base_reg, offset = c0
+    context.emit(Sh(c1, offset, base_reg))
+
+
+@isa.pattern('stm', 'STRU8(mem, reg)', size=2)
+@isa.pattern('stm', 'STRI8(mem, reg)', size=2)
+def pattern_sbi8(context, tree, c0, c1):
+    base_reg, offset = c0
+    context.emit(Sb(c1, offset, base_reg))
+
+
+@isa.pattern('reg', 'LDRI8(mem)', size=2)
 def pattern_ldri8(context, tree, c0):
     d = context.new_reg(RiscvRegister)
-    context.emit(Lb(d, 0, c0))
+    base_reg, offset = c0
+    context.emit(Lb(d, offset, base_reg))
     return d
 
+
+@isa.pattern('reg', 'LDRU8(mem)', size=2)
+def pattern_ldru8_fprel(context, tree, c0):
+    d = context.new_reg(RiscvRegister)
+    base_reg, offset = c0
+    context.emit(Lbu(d, offset, base_reg))
+    return d
+
+
+@isa.pattern('reg', 'LDRU32(mem)', size=2)
+@isa.pattern('reg', 'LDRI32(mem)', size=2)
+def pattern_ldr32_fprel(context, tree, c0):
+    d = context.new_reg(RiscvRegister)
+    base_reg, offset = c0
+    context.emit(Lw(d, offset, base_reg))
+    return d
+
+
+@isa.pattern('reg', 'NEGI8(reg)', size=2)
+@isa.pattern('reg', 'NEGI16(reg)', size=2)
+@isa.pattern('reg', 'NEGI32(reg)', size=2)
+def pattern_negi32(context, tree, c0):
+    context.emit(Subr(c0, R0, c0))
+    return c0
+
+
+@isa.pattern('reg', 'INVI8(reg)', size=2)
+@isa.pattern('reg', 'INVU8(reg)', size=2)
+@isa.pattern('reg', 'INVU32(reg)', size=2)
+@isa.pattern('reg', 'INVI32(reg)', size=2)
+def pattern_inv(context, tree, c0):
+    context.emit(Xori(c0, c0, -1))
+    return c0
+
+
 @isa.pattern('reg', 'LDRU16(reg)', size=2)
 def pattern_ldru16(context, tree, c0):
     d = context.new_reg(RiscvRegister)
     context.emit(Lhu(d, 0, c0))
     return d
 
+
 @isa.pattern('reg', 'LDRI16(reg)', size=2)
 def pattern_ldri16(context, tree, c0):
     d = context.new_reg(RiscvRegister)
     context.emit(Lh(d, 0, c0))
     return d
 
+
 @isa.pattern('reg', 'LDRU32(reg)', size=2)
 @isa.pattern('reg', 'LDRI32(reg)', size=2)
 def pattern_ldr_i32(context, tree, c0):
@@ -780,6 +901,8 @@ def pattern_ldr_i32(context, tree, c0):
     return d
 
 
+@isa.pattern('reg', 'ANDI8(reg, reg)', size=2)
+@isa.pattern('reg', 'ANDU8(reg, reg)', size=2)
 @isa.pattern('reg', 'ANDI16(reg, reg)', size=2)
 @isa.pattern('reg', 'ANDU16(reg, reg)', size=2)
 @isa.pattern('reg', 'ANDI32(reg, reg)', size=2)
@@ -813,18 +936,6 @@ def pattern_and8_reg_const(context, tree, c0):
     return d
 
 
-@isa.pattern(
-    'reg', 'ANDI8(CONSTI8, reg)', size=2,
-    condition=lambda t: t.children[0].value < 256)
-@isa.pattern(
-    'reg', 'ANDU8(CONSTU8, reg)', size=2,
-    condition=lambda t: t.children[0].value < 256)
-def pattern_and8_const_reg(context, tree, c0):
-    d = context.new_reg(RiscvRegister)
-    c1 = tree.children[0].value
-    context.emit(Andi(d, c0, c1))
-    return d
-
 @isa.pattern('reg', 'ORU32(reg, reg)', size=2)
 @isa.pattern('reg', 'ORI32(reg, reg)', size=2)
 @isa.pattern('reg', 'ORU16(reg, reg)', size=2)
@@ -857,33 +968,38 @@ def pattern_or_i32_const_reg(context, tree, c0):
     return d
 
 
-@isa.pattern('reg', 'SHRI32(reg, reg)', size=2)
+@isa.pattern('reg', 'SHRU8(reg, reg)', size=2)
+@isa.pattern('reg', 'SHRU16(reg, reg)', size=2)
 @isa.pattern('reg', 'SHRU32(reg, reg)', size=2)
-def pattern_shr_32(context, tree, c0, c1):
+def pattern_shr_u32(context, tree, c0, c1):
     d = context.new_reg(RiscvRegister)
     context.emit(Srl(d, c0, c1))
     return d
 
 
+@isa.pattern('reg', 'SHRI8(reg, reg)', size=2)
+@isa.pattern('reg', 'SHRI16(reg, reg)', size=2)
+@isa.pattern('reg', 'SHRI32(reg, reg)', size=2)
+def pattern_shr_i32(context, tree, c0, c1):
+    d = context.new_reg(RiscvRegister)
+    context.emit(Sra(d, c0, c1))
+    return d
+
+
 @isa.pattern(
     'reg', 'SHRI32(reg, CONSTI32)',
     size=2, condition=lambda t: t.children[1].value < 32)
 def pattern_shr_i32_reg_const(context, tree, c0):
     d = context.new_reg(RiscvRegister)
     c1 = tree.children[1].value
-    context.emit(Srli(d, c0, c1))
+    context.emit(Srai(d, c0, c1))
     return d
 
 
-@isa.pattern(
-    'reg', 'SHRI32(CONSTI32, reg)', size=2,
-    condition=lambda t: t.children[0].value < 32)
-def pattern_shr_i32_const_reg(context, tree, c0):
-    d = context.new_reg(RiscvRegister)
-    c1 = tree.children[0].value
-    context.emit(Srli(d, c0, c1))
-    return d
-
+@isa.pattern('reg', 'SHLU8(reg, reg)', size=2)
+@isa.pattern('reg', 'SHLI8(reg, reg)', size=2)
+@isa.pattern('reg', 'SHLU16(reg, reg)', size=2)
+@isa.pattern('reg', 'SHLI16(reg, reg)', size=2)
 @isa.pattern('reg', 'SHLU32(reg, reg)', size=2)
 @isa.pattern('reg', 'SHLI32(reg, reg)', size=2)
 def pattern_shl_i32(context, tree, c0, c1):
@@ -902,16 +1018,8 @@ def pattern_shl_i32_reg_const(context, tree, c0):
     return d
 
 
-@isa.pattern(
-    'reg', 'SHLI32(CONSTI32, reg)', size=2,
-    condition=lambda t: t.children[0].value < 32)
-def pattern_shl_i32_const_reg(context, tree, c0):
-    d = context.new_reg(RiscvRegister)
-    c1 = tree.children[0].value
-    context.emit(Slli(d, c0, c1))
-    return d
-
-
+@isa.pattern('reg', 'MULU8(reg, reg)', size=10)
+@isa.pattern('reg', 'MULU16(reg, reg)', size=10)
 @isa.pattern('reg', 'MULI32(reg, reg)', size=10)
 @isa.pattern('reg', 'MULU32(reg, reg)', size=10)
 def pattern_mul_i32(context, tree, c0, c1):
@@ -930,31 +1038,40 @@ def pattern_ldr_i32_add(context, tree, c0):
 
 
 @isa.pattern('reg', 'DIVI32(reg, reg)', size=10)
-@isa.pattern('reg', 'DIVU32(reg, reg)', size=10)
 def pattern_div_i32(context, tree, c0, c1):
     d = context.new_reg(RiscvRegister)
-    # Generate call into runtime lib function!
-    context.move(R12, c0)
-    context.move(R13, c1)
-    context.emit(Bl(LR, '__sdiv'))
-    context.move(d, R10)
+    context.emit(Div(d, c0, c1))
+    return d
+
+
+@isa.pattern('reg', 'DIVU16(reg, reg)', size=10)
+@isa.pattern('reg', 'DIVU32(reg, reg)', size=10)
+def pattern_div_u32(context, tree, c0, c1):
+    d = context.new_reg(RiscvRegister)
+    context.emit(Divu(d, c0, c1))
     return d
 
 
 @isa.pattern('reg', 'REMI32(reg, reg)', size=10)
 def pattern_rem_i32(context, tree, c0, c1):
-    # Implement remainder as a combo of div and mls (multiply substract)
     d = context.new_reg(RiscvRegister)
-    context.move(R12, c0)
-    context.move(R13, c1)
-    context.emit(Bl(LR, '__sdiv'))
-    context.move(d, R10)
-    context.emit(Mul(c1, c1, d))
-    d2 = context.new_reg(RiscvRegister)
-    context.emit(Subr(d2, c0, c1))
-    return d2
+    context.emit(Rem(d, c0, c1))
+    return d
 
 
+@isa.pattern('reg', 'REMU16(reg, reg)', size=10)
+@isa.pattern('reg', 'REMU32(reg, reg)', size=10)
+def pattern_rem_u32(context, tree, c0, c1):
+    d = context.new_reg(RiscvRegister)
+    context.emit(Remu(d, c0, c1))
+    return d
+
+
+@isa.pattern('reg', 'XORU8(reg, reg)', size=2)
+@isa.pattern('reg', 'XORI8(reg, reg)', size=2)
+@isa.pattern('reg', 'XORU16(reg, reg)', size=2)
+@isa.pattern('reg', 'XORI16(reg, reg)', size=2)
+@isa.pattern('reg', 'XORU32(reg, reg)', size=2)
 @isa.pattern('reg', 'XORI32(reg, reg)', size=2)
 def pattern_xor_i32(context, tree, c0, c1):
     d = context.new_reg(RiscvRegister)
