@@ -22,11 +22,11 @@ class CodeGenerator:
     """
     logger = logging.getLogger('pascal.codegen')
 
-    def __init__(self, diag, debug_db):
+    def __init__(self, diag):
         self.builder = irutils.Builder()
         self.diag = diag
         self.context = None
-        self.debug_db = debug_db
+        self.debug_db = debuginfo.DebugDb()
         self.module_ok = False
 
     def gencode(self, unit: nodes.Program, context):
@@ -37,7 +37,7 @@ class CodeGenerator:
         self.builder.prepare()
         self.module_ok = True
         self.logger.info('Generating ir-code for %s', unit.name)
-        self.builder.module = ir.Module(unit.name)
+        self.builder.module = ir.Module(unit.name, debug_db=self.debug_db)
 
         # Generate global variables:
         self.gen_globals(unit)
@@ -113,8 +113,9 @@ class CodeGenerator:
 
             # TODO
             size = 100  # elf.context.size_of(var.typ)
+            alignment = 4
             ir_var = ir.Variable(
-                var.name, size, value=cval)
+                var.name, size, alignment, value=cval)
             self.context.var_map[var] = ir_var
             self.builder.module.add_variable(ir_var)
 
@@ -219,6 +220,7 @@ class CodeGenerator:
             var_name = 'var_{}'.format(sym.name)
             variable = ir.Alloc(var_name, self.context.size_of(sym.typ))
             self.emit(variable)
+            variable = self.emit(ir.AddressOf(variable, var_name))
             if sym.isParameter:
                 # Get the parameter from earlier:
                 parameter = param_map[sym]
@@ -784,23 +786,25 @@ class CodeGenerator:
             if len(expr.val) == 1:
                 expr.typ = self.context.get_type('char')
                 val = ord(expr.val)
-                value = ir.Const(val, 'cnst', self.get_ir_type('char'))
+                value = self.emit(
+                    ir.Const(val, 'cnst', self.get_ir_type('char')))
             else:
                 expr.typ = self.context.get_type('string')
                 cval = self.context.pack_string(expr.val)
-                value = ir.LiteralData(cval, 'strval')
+                value = self.emit(ir.LiteralData(cval, 'strval'))
+                value = self.emit(ir.AddressOf(value, 'addr'))
         elif isinstance(expr.val, int):  # boolean is a subclass of int!
             # For booleans, use the integer as storage class:
             expr.typ = self.context.get_type('integer')
             val = int(expr.val)
-            value = ir.Const(val, 'cnst', self.get_ir_int())
+            value = self.emit(ir.Const(val, 'cnst', self.get_ir_int()))
         elif isinstance(expr.val, float):
             expr.typ = self.context.get_type('float')
             val = float(expr.val)
-            value = ir.Const(val, 'cnst', ir.f64)
+            value = self.emit(ir.Const(val, 'cnst', ir.f64))
         else:  # pragma: no cover
             raise NotImplementedError(str(expr.val))
-        return self.emit(value)
+        return value
 
     def gen_type_cast(self, expr):
         """ Generate code for type casting """
